package dev.fyloz.colorrecipesexplorer.logic.account

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.account.GroupTokenDto
import dev.fyloz.colorrecipesexplorer.dtos.account.GroupTokenSaveDto
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.logic.BaseLogic
import dev.fyloz.colorrecipesexplorer.service.account.GroupTokenService
import java.util.*
import javax.annotation.PostConstruct

interface GroupTokenLogic {
    fun isDisabled(id: String): Boolean
    fun getAll(): Collection<GroupTokenDto>
    fun getById(id: String): GroupTokenDto
    fun getById(id: UUID): GroupTokenDto
    fun save(dto: GroupTokenSaveDto): GroupTokenDto
    fun enable(id: String): GroupTokenDto
    fun disable(id: String): GroupTokenDto
    fun deleteById(id: String)
}

@LogicComponent
class DefaultGroupTokenLogic(private val service: GroupTokenService, private val groupLogic: GroupLogic) :
    GroupTokenLogic {
    private val typeName = Constants.ModelNames.GROUP_TOKEN
    private val typeNameLowerCase = typeName.lowercase()

    private val enabledTokensCache = hashSetOf<String>()

    @PostConstruct
    fun initEnabledTokensCache() {
        val tokensIds = getAll().filter { it.enabled }.map { it.id.toString() }
        enabledTokensCache.addAll(tokensIds)
    }

    override fun isDisabled(id: String) = !enabledTokensCache.contains(id)
    override fun getAll() = service.getAll()
    override fun getById(id: String) = getById(UUID.fromString(id))

    override fun getById(id: UUID) = service.getById(id) ?: throw notFoundException(value = id)

    override fun save(dto: GroupTokenSaveDto): GroupTokenDto {
        throwIfNameAlreadyExists(dto.name)

        val token = GroupTokenDto(
            generateUniqueUUIDForName(dto.name), dto.name, true, groupLogic.getById(dto.groupId)
        )

        val savedToken = service.save(token)
        enabledTokensCache.add(savedToken.id.toString())

        return savedToken
    }

    override fun enable(id: String) = setEnabled(id, true).also {
        enabledTokensCache.add(id)
    }

    override fun disable(id: String) = setEnabled(id, false).also {
        enabledTokensCache.remove(id)
    }

    override fun deleteById(id: String) {
        enabledTokensCache.remove(id)
        service.deleteById(UUID.fromString(id))
    }

    private fun setEnabled(id: String, enabled: Boolean) = with(getById(id)) {
        service.save(this.copy(enabled = enabled))
    }

    private fun generateUniqueUUIDForName(name: String): UUID {
        var id = generateUUIDForName(name)

        // UUIDs do not guarantee that collisions can't happen
        while (service.existsById(id)) {
            id = generateUUIDForName(name)
        }

        return id
    }

    private fun generateUUIDForName(name: String) = UUID.nameUUIDFromBytes(name.toByteArray())

    private fun throwIfNameAlreadyExists(name: String) {
        if (service.existsByName(name)) {
            throw alreadyExistsException(value = name)
        }
    }

    private fun notFoundException(identifierName: String = BaseLogic.ID_IDENTIFIER_NAME, value: Any) =
        NotFoundException(
            typeNameLowerCase,
            "$typeName not found",
            "A $typeNameLowerCase with the $identifierName '$value' could not be found",
            value,
            identifierName
        )

    private fun alreadyExistsException(identifierName: String = BaseLogic.NAME_IDENTIFIER_NAME, value: Any) =
        AlreadyExistsException(
            typeNameLowerCase,
            "$typeName already exists",
            "A $typeNameLowerCase with the $identifierName '$value' already exists",
            value,
            identifierName
        )
}
