package dev.fyloz.colorrecipesexplorer.logic.account

import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.dtos.account.UserDetails
import dev.fyloz.colorrecipesexplorer.dtos.account.UserJwt
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.utils.base64encode
import dev.fyloz.colorrecipesexplorer.utils.toDate
import io.jsonwebtoken.Jwts
import io.jsonwebtoken.jackson.io.JacksonDeserializer
import io.jsonwebtoken.jackson.io.JacksonSerializer
import org.springframework.stereotype.Service
import java.time.Instant
import java.util.*

const val jwtClaimUser = "user"

interface JwtLogic {
    /** Build a JWT for the given [userDetails]. */
    fun buildUserJwt(userDetails: UserDetails): String

    /** Build a JWT for the given [groupTokenId]. */
    fun buildGroupTokenIdJwt(groupTokenId: UUID): String

    /** Parses a user information from the given [jwt]. */
    fun parseUserJwt(jwt: String): UserJwt

    /** Parses a group token id from the given [jwt]. */
    fun parseGroupTokenIdJwt(jwt: String): UUID
}

@Service
class DefaultJwtLogic(
    val objectMapper: ObjectMapper,
    val securityProperties: CreSecurityProperties
) : JwtLogic {
    private val secretKey by lazy {
        securityProperties.jwtSecret.base64encode()
    }

    private val permissionsById = Permission.values()
        .associateBy { it.id }

    // Must be a new instance every time, or data from the last token will still be there
    private val jwtBuilder
        get() =
            Jwts.builder()
                .serializeToJsonWith(JacksonSerializer<Map<String, *>>(objectMapper))
                .signWith(secretKey)

    private val jwtParser by lazy {
        Jwts.parserBuilder()
            .deserializeJsonWith(JacksonDeserializer<Map<String, *>>(objectMapper))
            .setSigningKey(secretKey)
            .build()
    }

    override fun buildUserJwt(userDetails: UserDetails): String {
        val permissionsIds = userDetails.permissions.map { it.id }
        val type = if (userDetails.isGroup) JWT_TYPE_GROUP else JWT_TYPE_USER

        return jwtBuilder
            .setSubject(userDetails.id)
            .setExpiration(getCurrentExpirationDate())
            .claim(JWT_CLAIM_PERMISSIONS, objectMapper.writeValueAsString(permissionsIds))
            .claim(JWT_CLAIM_TYPE, type)
            .compact()
    }

    override fun buildGroupTokenIdJwt(groupTokenId: UUID): String =
        jwtBuilder
            .setSubject(groupTokenId.toString())
            .compact()

    override fun parseUserJwt(jwt: String): UserJwt {
        val parsedJwt = jwtParser.parseClaimsJws(jwt)

        val serializedPermissions = parsedJwt.body.get(JWT_CLAIM_PERMISSIONS, String::class.java)
        val permissionsIds = objectMapper.readValue<Collection<Int>>(serializedPermissions)
        val permissions = permissionsIds.map { permissionsById[it]!! }

        val type = parsedJwt.body[JWT_CLAIM_TYPE] as Int
        val isGroup = type == JWT_TYPE_GROUP

        val authorities = permissions
            .map { it.toAuthority() }
            .toMutableList()

        return UserJwt(parsedJwt.body.subject, authorities, isGroup)
    }

    override fun parseGroupTokenIdJwt(jwt: String): UUID {
        val uuid = jwtParser.parseClaimsJws(jwt).body.subject
        return UUID.fromString(uuid)
    }

    private fun getCurrentExpirationDate(): Date =
        Instant.now()
            .plusSeconds(securityProperties.jwtDuration)
            .toDate()

    companion object {
        private const val JWT_CLAIM_PERMISSIONS = "perms"
        private const val JWT_CLAIM_TYPE = "type"

        private const val JWT_TYPE_USER = 0
        private const val JWT_TYPE_GROUP = 1
    }
}
