package dev.fyloz.colorrecipesexplorer.service.account

import dev.fyloz.colorrecipesexplorer.config.annotations.ServiceComponent
import dev.fyloz.colorrecipesexplorer.dtos.account.GroupDto
import dev.fyloz.colorrecipesexplorer.dtos.account.UserDto
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.model.account.User
import dev.fyloz.colorrecipesexplorer.repository.UserRepository
import dev.fyloz.colorrecipesexplorer.service.BaseService
import dev.fyloz.colorrecipesexplorer.service.Service

interface UserService : Service<UserDto, User, UserRepository> {
    /** Checks if a user with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String, id: Long? = null): Boolean

    /** Gets all users, depending on [isSystemUser]. */
    fun getAll(isSystemUser: Boolean): Collection<UserDto>

    /** Gets all users for the given [group]. */
    fun getAllByGroup(group: GroupDto): Collection<UserDto>

    /** Finds the user with the given [id], depending on [isSystemUser]. */
    fun getById(id: Long, isSystemUser: Boolean): UserDto?

    /** Finds the user with the given [firstName] and [lastName]. */
    fun getByFirstNameAndLastName(firstName: String, lastName: String): UserDto?
}

@ServiceComponent
class DefaultUserService(repository: UserRepository, private val groupService: GroupService) :
    BaseService<UserDto, User, UserRepository>(repository), UserService {
    override fun existsByFirstNameAndLastName(firstName: String, lastName: String, id: Long?) =
        repository.existsByFirstNameAndLastNameAndIdNotAndIsDefaultGroupUserIsFalse(firstName, lastName, id ?: 0L)

    override fun getAll(isSystemUser: Boolean) =
        repository.findAllByIsDefaultGroupUserIsFalse()
            .filter { isSystemUser || !it.isSystemUser }
            .map(::toDto)

    override fun getAllByGroup(group: GroupDto) =
        repository.findAllByGroup(groupService.toEntity(group))
            .map(::toDto)

    override fun getById(id: Long, isSystemUser: Boolean): UserDto? {
        val user = repository.findByIdAndIsDefaultGroupUserIsFalse(id) ?: return null
        if (!isSystemUser && user.isSystemUser) {
            return null
        }

        return toDto(user)
    }

    override fun getByFirstNameAndLastName(firstName: String, lastName: String): UserDto? {
        val user = repository.findByFirstNameAndLastName(firstName, lastName)
        return if (user != null) toDto(user) else null
    }

    override fun toDto(entity: User) = UserDto(
        entity.id,
        entity.firstName,
        entity.lastName,
        entity.password,
        if (entity.group != null) groupService.toDto(entity.group) else null,
        getFlattenPermissions(entity),
        entity.permissions,
        entity.lastLoginTime,
        entity.isSystemUser
    )

    override fun toEntity(dto: UserDto) = User(
        dto.id,
        dto.firstName,
        dto.lastName,
        dto.password,
        false,
        dto.isSystemUser,
        if (dto.group != null) groupService.toEntity(dto.group) else null,
        dto.explicitPermissions,
        dto.lastLoginTime
    )

    private fun getFlattenPermissions(user: User): List<Permission> {
        val perms = user.permissions.flatMap { it.flat() }.filter { !it.deprecated }

        if (user.group != null) {
            return perms + groupService.flattenPermissions(user.group)
        }

        return perms
    }
}
